from genshi.builder import tag
from genshi.filters import Transformer

from trac.core import implements, Component
from trac.ticket.api import ITicketActionController
from trac.ticket.default_workflow import ConfigurableTicketWorkflow
from trac.util import get_reporter_id
from trac.web.chrome import add_warning, Chrome
from trac.web.api import ITemplateStreamFilter

class CCHandlerActionController(Component):
    """ Automatically CC the User based on his/her selection of a field on the screen

    Don't forget to add `CCHandlerActionController` to the workflow
    option in [ticket].
    If there is no workflow option, the line will look like this:

    workflow = ConfigurableTicketWorkflow, CCHandlerActionController
    """
    implements(ITemplateStreamFilter, ITicketActionController)

    def __init__(self):
        self.workflow = ConfigurableTicketWorkflow(self.env)

    # ITemplateStreamFilter methods

    def filter_stream(self, req, method, filename, stream, data):
        if filename == 'ticket.html' and req.authname != 'anonymous':
            ticket = data.get('ticket')
            if ticket and ticket.exists:
                #filter = Transformer("div[@class='field']/fieldset[@class='iefix']")
                #filter = Transformer("div[@id='content' and @class='ticket']/div[@id='ticket']")
                filter = Transformer("//form[@id='propertyform']/div[@class='field']")
                #Hide the CC checkbox if it is present
                if data.has_key('fields'):
                  for dict in data['fields']:
                    if dict['name'] == 'cc':
                      if dict.has_key('cc_entry'):
                        dict['skip'] = True
                        dict['rendered'] = 'no'
                      break


                return stream | filter.before(self._append_CC_radio(req, ticket)) 
        return stream

    # ITicketActionController methods

    def get_ticket_actions(self, req, ticket):
        #We need to handle all statuses so that we can perform the action regardless of which action was selected
        #TODO: What about other actions not in ConfigurableTicketWorkflow? (We'll worry about it when it happens)
        return self.workflow.get_ticket_actions(req, ticket)

    def get_all_status(self):
        return []

    def render_ticket_action_control(self, req, ticket, action):
        return (None, None, None)

    def get_ticket_changes(self, req, ticket, action):
        return {}

    def apply_action_side_effects(self, req, ticket, action):
        if (action != 'delete' and 'cc_update_manual' in req.args):
            cc_action, cc_entry, cc_list = self._tweak_cc(req, ticket['cc'], req.args['cc_update_manual'] == 'yes')

            if cc_action == 'remove':
                self.log.info("Removing CC %s on ticket %s" % (cc_entry, ticket.id))
                #add_warning(req, "You have been removed from the Cc list.")
                cc_list.remove(cc_entry)
            elif cc_action == 'add':
                self.log.info("Adding CC %s on ticket %s" % (cc_entry, ticket.id))
                #add_warning(req, "You have been added to the Cc list.")
                cc_list.append(cc_entry)
            else:
                return

            new_cc_val = ', '.join(cc_list)

            with self.env.db_transaction as db:
                cursor = db.cursor()
                cursor.execute("UPDATE ticket SET cc=%s WHERE id=%s",
                           (new_cc_val, ticket.id))

    #Internal Functions

    def _append_CC_radio(self, req, ticket):
        notify_reporter = self.env.config.getbool('notification',
                                                  'always_notify_reporter')
        notify_owner = self.env.config.getbool('notification',
                                               'always_notify_owner')
        notify_updater = self.env.config.getbool('notification',
                                                 'always_notify_updater')

        self.log.info("Appending CC Radio Button to ticket %s for user %s" % (ticket.id, req.authname))

        add_checked = False
        remove_checked = False
        disabled = False
        notemsg = None

        if notify_reporter and ticket.values.get('reporter') == req.authname:
            notemsg = " (You are permanently added to the cc list because you are the reporter.)"
            add_checked = True
            disabled = True
        elif notify_owner and ticket.values.get('owner') == req.authname:
            notemsg = " (You are automatically added to the cc list because you are the owner.)"
            add_checked = True
            disabled = True
        elif notify_updater:
            if self._already_changed_ticket(req, ticket):
                add_checked = True
                disabled = True
                notemsg = " (You are permanently added to the cc list because you commented on the ticket.)"
            else:
                notemsg = " Note: Updating the ticket will permanently add you to the cc list."

        if not add_checked:
            add_checked = 'cc_update_manual' in req.args and req.args['cc_update_manual'] == 'yes'
            remove_checked = 'cc_update_manual' in req.args and req.args['cc_update_manual'] == 'no'
            #If this isn't a preview with a value already selected, select the current
            if (not add_checked and not remove_checked):
                cc_action, cc_entry, cc_list = self._tweak_cc(req, ticket['cc'], True)
                if cc_action != 'add':
                    add_checked = True
                else:
                    remove_checked = True

        return tag.p("Cc me to this ticket: ",
            tag.input(type="radio", id="field-cc-manual1", name="cc_update_manual",
                      value="yes", checked=(add_checked or None), disabled=(disabled or None)),
            tag.label("Yes", for_="field-cc-manual1"), " ",
            tag.input(type="radio", id="field-cc-manual0", name="cc_update_manual",
                      value="no", checked=(remove_checked or None), disabled=(disabled or None)),
            tag.label("No", for_="field-cc-manual0"),
            notemsg and tag.em(notemsg) or "")

    def _already_changed_ticket(self, req, ticket):
        with self.env.db_query as db: 
            cursor = db.cursor()
            cursor.execute("SELECT DISTINCT author,ticket FROM ticket_change "
                       "WHERE ticket=%s AND author=%s", (ticket.id, req.authname))
            for author,ticket in cursor:
                return True

        return False

    def _tweak_cc(self, req, cc, cc_me):
        """Return an (action, recipient, cc_list) tuple corresponding to a change
        of CC status for this user relative to the current `cc_list`."""
        entries = []
        email = req.session.get('email', '').strip()
        if req.authname != 'anonymous':
            entries.append(req.authname)
        elif email:
            entries.append(email)
        else:
            author = get_reporter_id(req, 'author').strip()
            if author and author != 'anonymous':
                email = author.split()[-1]
                if (email[0], email[-1]) == ('<', '>'):
                    email = email[1:-1]
                entries.append(email)

        add = []
        remove = []
        cc_list = Chrome(self.env).cc_list(cc)
        for entry in entries:
            if entry in cc_list:
                if not cc_me:
                    remove.append(entry)
            else:
                if cc_me:
                    add.append(entry)

        action = entry = ''
        if remove:
            action, entry = ('remove', remove[0])
        elif add:
            action, entry = ('add', add[0])
        return (action, entry, cc_list)

